import asyncio
import functools
import typing

try:
    import contextvars # python 3.7
except ImportError:
    contextvars = None


T = typing.TypeVar("T")


async def run_in_threadpool(function: typing.Callable[..., T], *args: typing.Any, **kwargs: typing.Any) -> T:
    loop = asyncio.get_event_loop()

    if contextvars is not None:
        child = functools.partial(function, *args, **kwargs)
        context = contextvars.copy_context()
        function = context.run
        args = (child,)
    elif kwargs:
        function = functools.partial(function, **kwargs)

    return await loop.run_in_executor(None, function, *args)



import http

from .responses import PlainTextResponse


class HttpException(Exception):
    def __init__(self, status_code: int, details: str = None) -> None:
        if details is None:
            details = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.details = details

    @property
    def response(self) -> PlainTextResponse:
        return PlainTextResponse(
            status_code=self.status_code,
            content=self.details,
        )

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.details!r})"

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.details!r})"



from parse import parse
import typing

from .constants import HTTP_METHODS


class Route:
    def __init__(self, path: str, handler: callable, methods: list = None, is_static: bool = False):
        self.path = path
        self.handler = handler
        self.methods = methods if methods else HTTP_METHODS
        self.is_static = is_static  # check if it is static route

    @property
    def path(self) -> str:
        return self.__path

    @path.setter
    def path(self, path: str) -> None:
        self.__path = path

    @property
    def handler(self) -> callable:
        return self.__handler

    @handler.setter
    def handler(self, handler: callable) -> None:
        self.__handler = handler

    @property
    def methods(self) -> list:
        return self.__methods

    @methods.setter
    def methods(self, methods: list) -> None:
        # make sure all HTTP methods are upper
        self.__methods = [method.upper() for method in methods]

    def is_valid_method(self, method: str) -> bool:
        return method.upper() in self.methods


class Router:
    def __init__(self):
        self.routes = []

    @property
    def routes(self) -> typing.List[Route]:
        return self.__routes

    @routes.setter
    def routes(self, routes: typing.List[Route]):
        self.__routes = routes

    @property
    def paths(self) -> typing.List[str]:
        paths = set()
        for route in self.routes:
            paths.add(route.path)
        return list(paths)

    def add_route(self, path: str, handler: callable, methods: list = None, is_static: bool = False) -> None:
        assert path not in self.paths, f"Route {path}, already exists"
        self.routes.append(
            Route(path=path, handler=handler, methods=methods, is_static=is_static)
        )

    def route(self, path: str, methods: list = None) -> callable:
        def wrapper(handler):
            self.add_route(path, handler, methods)
            return handler

        return wrapper

    def mount(self, router: callable, prefix: str = None, is_static: bool = False) -> None:
        """Mount another router"""
        routes = router.routes

        # register sub routes
        for route in routes:
            path = self.__add_prefix(prefix, route.path) if prefix else route.path
            self.add_route(path=path, handler=route.handler, methods=route.methods, is_static=is_static)

    def get_route(self, request_path: str, method: str) -> (Route, typing.Any):
        for route in self.routes:
            # for static routing, use different methods for route comparison
            if route.is_static and request_path.startswith(route.path):
                return route, {}

            parse_result = parse(route.path, request_path)
            if parse_result is not None:
                return route, parse_result.named
    
        return None, None

    def __add_prefix(self, prefix: str, path: str) -> str:
        if not prefix.startswith("/"):
            prefix = "/" + prefix

        return f"{prefix}{path}" if path.startswith("/") else f"{prefix}/{path}"



import httpx
import inspect
import typing

from .exceptions import HttpException
from .middleware import BaseMiddleware, ExceptionMiddleware
from .requests import Request
from .responses import Response, FileResponse
from .routing import Router
from .staticfiles import StaticFiles
from .types import Scope, Receive, Send


class Yaat:
    def __init__(self):
        self.router = Router()
        self.middleware = BaseMiddleware(self)
        self.exception_handler = None

        # register exception handling middleware
        self.add_middleware(ExceptionMiddleware)

    # NOTE: properties
    @property
    def exception_handler(self) -> callable:
        return self.__exception_handler

    @exception_handler.setter
    def exception_handler(self, exception: callable) -> None:
        self.__exception_handler = exception


    # NOTE: Routing
    def add_route(self, path: str, handler: callable, methods: list = None) -> None:
        self.router.add_route(path, handler, methods)

    def route(self, path: str, methods: list = None) -> callable:
        def wrapper(handler):
            self.add_route(path, handler, methods)
            return handler

        return wrapper

    def mount(self, router: Router, prefix: str = None) -> None:
        # check if its static route
        is_static = isinstance(router, StaticFiles)

        if prefix and is_static:
            # NOTE: because 'prefix' is already defined in static route
            raise ValueError("'prefix' must be None when mounting static routes.")

        self.router.mount(
            router=router,
            prefix=prefix,
            is_static=is_static,
        )


    # NOTE: Handle Request
    async def handle_request(self, request: Request) -> Response:
        route, kwargs = self.router.get_route(request_path=request.path, method=request.method)

        try:
            if route and route.handler is not None:
                handler = route.handler

                if inspect.isclass(handler):
                    handler = getattr(handler(), request.method.lower(), None)
                    if handler is None:
                        raise HttpException(405)
                if not route.is_valid_method(request.method):
                    raise HttpException(405)

                response = await handler(request, **kwargs)
            else:
                # default response when path not found
                raise HttpException(404)
        except Exception as e:
            if self.exception_handler is not None:
                response = self.exception_handler(request, e)
            elif isinstance(e, HttpException) or isinstance(e, HttpException):
                response = e.response
            else:
                raise e
        return response


    # NOTE: Middleware
    def add_middleware(self, middleware_cls: BaseMiddleware) -> None:
        self.middleware.add(middleware_cls)


    # NOTE: Test Client
    def test_client(self, base_url="http://testserver") -> httpx.AsyncClient:
        if not hasattr(self, "_session"):
            self._session = httpx.AsyncClient(app=self, base_url="http://testserver")

        return self._session


    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.middleware(scope, receive, send)



import typing

from ..requests import Request
from ..responses import Response
from ..types import ASGIApp, Scope, Receive, Send


class BaseMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    def add(self, middleware_cls: typing.Type, **kwargs):
        self.app = middleware_cls(self.app, **kwargs)

    async def process_request(self, request: Request):
        pass

    async def process_response(self, response: Response):
        pass

    async def handle_request(self, request: Request) -> Response:
        await self.process_request(request)
        response = await self.app.handle_request(request)
        await self.process_response(response)
        return response

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        response = await self.handle_request(request)
        await response(send)



from email.utils import parsedate
import os
import typing

try:
    from aiofiles.os import stat as aio_stat
except ImportError:
    aio_stat = None

from .exceptions import HttpException
from .requests import Request
from .responses import FileResponse, NotModifiedResponse, Response
from .routing import Router, Route


class StaticFilesHandler:
    def __init__(self, path: str = "/", directory: str = "", html: bool = False):
        self.path = path
        self.directory = directory
        self.html = html

    @property
    def directory(self) -> str:
        return self.__directory

    @directory.setter
    def directory(self, directory: str):
        if os.path.isfile(directory):
            raise RuntimeError(f"StaticFiles directory {directory} is not a directory.")
        self.__directory = directory

    def is_not_modified(self, request_headers: dict, response_headers: dict) -> bool:
        """
        Check if file is modified, if not return "Not Modified" response instead.
        """

        try:
            if_none_match = request_headers["if-none-match"]
            etag = response_headers["etag"]
            if if_none_match == etag:
                return True
        except KeyError:
            pass

        try:
            if_modified_since = parsedate(request_headers["if-modified-since"])
            last_modified = parsedate(response_headers["last-modified"])
            if (
                if_modified_since is not None
                and last_modified is not None
                and if_modified_since >= last_modified
            ):
                return True
        except KeyError:
            pass

        return False

    async def __call__(self, request: Request, *args, **kwargs) -> Response:
        assert aio_stat is not None, "'aiofiles' must be installed to use StaticFilesHandler"

        request_path = request.path
        # NOTE: remove route prefix and get file path
        if request_path.startswith(self.path) and self.path != "/":
            filepath = request_path[len(self.path):]
        else:
            filepath = request_path
        # if starts with / remove
        filepath = filepath[1:] if filepath.startswith("/") else filepath

        try:
            full_path = os.path.join(self.directory, filepath)
            is_directory = os.path.isdir(full_path)
            is_file_exists = os.path.exists(full_path)

            # if html response
            if self.html:
                index_path = os.path.join(self.directory, "index.html")
                is_index_exists = os.path.exists(index_path)

                # if directory, check if there is index file
                if is_index_exists and is_directory:
                    full_path = index_path
                # raise 404 if not directory and file also not exists
                elif not is_file_exists and not is_directory:
                    raise HttpException(404)

                stat_result = await aio_stat(full_path)
                response = FileResponse(path=full_path, stat_result=stat_result)

            # if file response
            else:
                if is_directory or not is_file_exists:
                    raise HttpException(
                        status_code=404,
                        details="File does not exists"
                    )

                stat_result = await aio_stat(full_path)
                response = FileResponse(path=full_path, stat_result=stat_result)

                # check if file is modified
                if self.is_not_modified(dict(request.headers), response.headers):
                    response = NotModifiedResponse(response.headers)

        except HttpException as e:
            response = e.response

        return response


class StaticFiles:
    def __init__(self, path: str, directory: str, html: bool = False):
        self.path = path
        self.router = Router()
        self.router.add_route(
            path=self.path,
            handler=StaticFilesHandler(self.path, directory, html),
            methods=["GET", "HEAD"]
        )

    @property
    def path(self) -> str:
        return self.__path

    @path.setter
    def path(self, path: str) -> None:
        # clean static directory path
        if path.endswith("/"):
            path = path[:-1]
        if not path.startswith("/"):
            path = f"/{path}"

        self.__path = path

    @property
    def routes(self) -> typing.List[Route]:
        return self.router.routes



from urllib.parse import parse_qsl, urlparse
import tempfile
import typing

from .constants import ENCODING_METHOD
from .concurrency import run_in_threadpool
from .types import Scope


class URL:
    def __init__(self, url: str = "", scope: Scope = None):
        if scope is not None:
            assert not url, 'Cannot set both "url" and "scope".'

            self.scheme = scope.get("scheme", "http")
            self.server = scope.get("server", None)  
            self.path = scope.get("root_path", "") + scope["path"]
            self.query = scope.get("query_string", b"")
            self.host_header = scope["headers"]
        else:
            assert not scope, 'Cannot set both "url" and "scope".'

            components = urlparse(url)
            netloc = components.netloc.split(":")

            self.scheme = components.scheme
            self.server = (netloc[0], netloc[1] if len(netloc) > 1 else None)
            self.path = components.path
            self.query = components.query
            self.host_header = []
            self.fragment = components.fragment

        # init url
        self.__init_url()

    @property
    def scheme(self) -> str:
        return self.__scheme

    @scheme.setter
    def scheme(self, scheme: str):
        self.__scheme = scheme

    @property
    def server(self) -> typing.Tuple[str, int]:
        return self.__server

    @server.setter
    def server(self, server: typing.Tuple[str, int]):
        self.__server = server # ip and port

    @property
    def netloc(self) -> str:
        if self.port:
            return f"{self.host}:{self.port}"
        else:
            return self.host

    @property
    def host(self) -> str:
        if not hasattr(self, "_host"):
            self.__host = None

            if self.server:
                host, port = self.server
                self._host = host
        return self._host

    @property
    def port(self) -> int:
        if not hasattr(self, "_port"):
            self._port = None

            if self.server:
                host, port = self.server
                self._port = int(port) if port else None
        return self._port

    @property
    def query(self) -> str:
        return self.__query

    @query.setter
    def query(self, query: bytes):
        try:
            self.__query = query.decode(ENCODING_METHOD)
        except (AttributeError, UnicodeDecodeError):
            self.__query = query

    @property
    def host_header(self) -> str:
        return self.__host_header

    @host_header.setter
    def host_header(self, headers: list):
        host_header = None
        for key, value in headers:
            if key == b"host":
                host_header = value.decode(ENCODING_METHOD)
                break
        self.__host_header = host_header

    @property
    def fragment(self) -> str:
        return self.__fragment

    @fragment.setter
    def fragment(self, fragment: str):
        self.__fragment = fragment

    @property
    def url(self) -> str:
        return self.__url

    def __init_url(self):
        if self.host_header is not None:
            url = f"{self.scheme}://{self.host_header}{self.path}"
        elif self.server is None:
            url = self.path
        else:
            host, port = self.server
            default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[self.scheme]
            if port == default_port:
                url = f"{self.scheme}://{host}{self.path}"
            else:
                url = f"{self.scheme}://{host}:{port}{self.path}"

        if self.query:
            url += "?" + self.query

        self.__url = url

    def is_secure(self) -> bool:
        return self.scheme in ("https", "wss")

    def __eq__(self, other: typing.Any) -> bool:
        return str(self) == str(other)

    def __str__(self) -> str:
        return self.__url


class Address:
    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port

    @property
    def host(self) -> str:
        return self.__host

    @host.setter
    def host(self, host: str):
        self.__host = str(host)

    @property
    def port(self) -> int:
        return self.__port

    @port.setter
    def port(self, port: int):
        self.__port = int(port) if port else None

    def __str__(self) -> str:
        if self.host and self.port:
            return f"{self.host}:{self.port}"
        return self.host


class Headers:
    def __init__(self, raw_headers: typing.List[typing.Tuple[bytes, bytes]]):
        self.__raw_headers = raw_headers
        self.__init_headers()

    @property
    def raw(self) -> typing.List[typing.Tuple[bytes, bytes]]:
        return self._raw_headers

    def __init_headers(self):
        self.__headers = {key.decode(ENCODING_METHOD): value.decode(ENCODING_METHOD) for key, value in self.__raw_headers}

    def items(self) -> dict:
        return self.__headers.items()

    def get(self, key: str, default: typing.Any = None) -> str:
        try:
            return self.__headers[key]
        except KeyError:
            return default

    def __contains__(self, key: typing.Any) -> bool:
        return key in self.__headers

    def __getitems__(self, key: typing.Any) -> str:
        return self.__headers[key]

    def __iter__(self) -> typing.Iterator[typing.Any]:
        for key, value in self.items():
            yield key, value

    def __len__(self) -> int:
        return len(self.__headers)


class QueryParams:
    def __init__(self, raw_query: bytes):
        self.__raw_query = raw_query
        self.__init_query()

    @property
    def raw(self) -> bytes:
        return self.__raw_query

    def __init_query(self):
        try:
            query = self.raw.decode(ENCODING_METHOD)
        except (AttributeError, UnicodeDecodeError):
            query = self.raw

        self.__query = self.__format_to_dict(parse_qsl(query, keep_blank_values=True))

    def __format_to_dict(self, query_list: list):
        query = {}

        for qs in query_list:
            key = qs[0]
            value = qs[1]

            if not key in query:
                query[key] = value
                continue

            # if key exists, store multiple values in list
            values = query[key]
            # convert to list if not a list already
            if type(values) != list:
                values = [values]
            values.append(value)
            query[key] = values

        return query

    def items(self) -> dict:
        return self.__query.items()

    def get(self, key: str, default: typing.Any = None) -> str:
        try:
            return self.__query[key]
        except KeyError:
            return default

    def __contains__(self, key: typing.Any) -> bool:
        return key in self.__query

    def __getitems__(self, key: typing.Any) -> str:
        return self.__query[key]

    def __iter__(self) -> typing.Iterator[typing.Any]:
        for key, value in self.items():
            yield key, value

    def __len__(self) -> int:
        return len(self.__query)

    def __str__(self) -> str:
        query_string = None

        for key, value in self.items():
            # if no value, just add back key
            if not value:
                query_string = f"{key}" if not query_string else f"{query_string}&{key}"
            # if list, iterate and add back
            elif type(value) == list:
                for each in value:
                    query_string = f"{key}={each}" if not query_string else\
                        f"{query_string}&{key}={each}"
            # just add
            else:
                query_string = f"{key}={value}" if not query_string else\
                    f"{query_string}&{key}={value}"

        return query_string


class Form:
    def __init__(self, form_data: typing.List[typing.Tuple[str, str]] = None):
        self.raw = form_data if form_data else []

    @property
    def raw(self):
        return self.__raw

    @raw.setter
    def raw(self, form_data: typing.List[typing.Tuple[str, str]]) -> typing.List:
        self.__raw = form_data
        self.__init_data()

    def __init_data(self):
        data = {}
        for item in self.raw:
            key = item[0]
            value = item[1]

            if not key in data:
                data[key] = value
                continue

             # if key exists, store multiple values in list
            values = data[key]
            # convert to list if not a list already
            if type(values) != list:
                values = [values]
            values.append(value)
            data[key] = values

        self.__data = data

    def items(self) -> typing.Dict:
        return self.__data.items()

    def get(self, key: str, default: typing.Any = None) -> typing.Any:
        return self.__data.get(key, default)

    def __contains__(self, key: typing.Any) -> bool:
        return key in self.__data

    def __getitems__(self, key: typing.Any) -> str:
        return self.__data[key]

    def __iter__(self) -> typing.Iterator[typing.Any]:
        for key, value in self.items():
            yield key, value

    def __len__(self) -> int:
        return len(self.__data)


class UploadFile:
    SPOOL_MAX_SIZE = 1024 * 1024

    def __init__(self, name: str, file: typing.IO = None, content_type: str = "") -> None:
        self.name = name
        self.content_type = content_type
        if file is None:
            file = tempfile.SpooledTemporaryFile(max_size=self.SPOOL_MAX_SIZE)
        self.file = file

    async def write(self, data: typing.Union[bytes, str]) -> None:
        await run_in_threadpool(self.file.write, data)

    async def read(self, size: int = None) -> typing.Union[bytes, str]:
        return await run_in_threadpool(self.file.read, size)

    async def seek(self, offset: int) -> None:
        await run_in_threadpool(self.file.seek, offset)

    async def close(self) -> None:
        await run_in_threadpool(self.file.close)



import traceback

from .base import BaseMiddleware
from ..exceptions import HttpException
from ..requests import Request
from ..responses import Response


class ExceptionMiddleware(BaseMiddleware):
    async def handle_request(self, request: Request) -> Response:
        await self.process_request(request)

        try:
            response = await self.app.handle_request(request)
        except Exception as err:
            # print traceback and send 500 Internal Server Error
            traceback.print_exc()
            response = HttpException(500).response        

        await self.process_response(response)
        return response



from urllib.parse import parse_qsl
from enum import Enum
import typing

try:
    from multipart.multipart import parse_options_header
    import multipart
except ImportError:  # pragma: nocover
    parse_options_header = None
    multipart = None

from .constants import ENCODING_METHOD
from .datatypes import Form, Headers, UploadFile


class FormParser:
    def __init__(self, body: bytes):
        self.body = body

    async def parse(self) -> Form:
        body_data = await self.body()
        body_data = body_data.decode(ENCODING_METHOD)
        form_data = [] if body_data == '' else parse_qsl(body_data)
        form_data = [(item[0], item[1]) for item in form_data]
        return Form(form_data)


class MultiPartParser:
    class MultiPartMessage(Enum):
        PART_BEGIN = 1
        PART_DATA = 2
        PART_END = 3
        HEADER_FIELD = 4
        HEADER_VALUE = 5
        HEADER_END = 6
        HEADERS_FINISHED = 7
        END = 8


    def __init__(self, headers: Headers, stream: typing.AsyncGenerator[bytes, None]):
        self.headers = headers
        self.stream = stream
        self.messages = []

    def __user_safe_decode(self, src: bytes, codec: str) -> str:
        try:
            return src.decode(codec)
        except (UnicodeDecodeError, LookupError):
            return src.decode(ENCODING_METHOD)

    async def parse(self) -> Form:
        events = self.MultiPartMessage

        # Parse the Content-Type header to get the multipart boundary.
        content_type, params = parse_options_header(self.headers.get("content-type"))
        charset = params.get(b"charset", "utf-8")
        if type(charset) == bytes:
            charset = charset.decode(ENCODING_METHOD)
        boundary = params.get(b"boundary")

        # Callbacks dictionary.
        callbacks = {
            "on_part_begin": lambda: self.messages.append((events.PART_BEGIN, b"")),
            "on_part_data": lambda data, start, end: self.messages.append((events.PART_DATA, data[start:end])),
            "on_part_end": lambda: self.messages.append((events.PART_END, b"")),
            "on_header_field": lambda data, start, end: self.messages.append((events.HEADER_FIELD, data[start:end])),
            "on_header_value": lambda data, start, end: self.messages.append((events.HEADER_VALUE, data[start:end])),
            "on_header_end": lambda: self.messages.append((events.HEADER_END, b"")),
            "on_headers_finished": lambda: self.messages.append((events.HEADERS_FINISHED, b"")),
            "on_end": lambda: self.messages.append((events.END, b"")),
        }

        # Create the parser.
        parser = multipart.MultipartParser(boundary, callbacks)
        header_field = b""
        header_value = b""
        content_disposition = None
        content_type = b""
        field_name = ""
        data = b""
        file = None

        items = (
            []  # type: typing.List[typing.Tuple[str, typing.Union[str, UploadFile]]]
        )

        # Data from the request
        async for chunk in self.stream:
            parser.write(chunk)
            messages = list(self.messages)
            self.messages.clear()

            for message_type, message_bytes in messages:
                if message_type == events.PART_BEGIN:
                    content_disposition = None
                    content_type = b""
                    data = b""

                elif message_type == events.HEADER_FIELD:
                    header_field += message_bytes

                elif message_type == events.HEADER_VALUE:
                    header_value += message_bytes

                elif message_type == events.HEADER_END:
                    field = header_field.lower()

                    if field == b"content-disposition":
                        content_disposition = header_value
                    elif field == b"content-type":
                        content_type = header_value

                    header_field = b""
                    header_value = b""

                elif message_type == events.HEADERS_FINISHED:
                    disposition, options = parse_options_header(content_disposition)
                    field_name = self.__user_safe_decode(options[b"name"], charset)

                    if b"filename" in options:
                        filename = self.__user_safe_decode(options[b"filename"], charset)
                        file = UploadFile(
                            name=filename,
                            content_type=content_type.decode(ENCODING_METHOD),
                        )
                    else:
                        file = None

                elif message_type == events.PART_DATA:
                    if file is None:
                        data += message_bytes
                    else:
                        await file.write(message_bytes)

                elif message_type == events.PART_END:
                    if file is None:
                        items.append((field_name, self.__user_safe_decode(data, charset)))
                    else:
                        await file.seek(0)
                        items.append((field_name, file))

                elif message_type == events.END:
                    pass

        parser.finalize()
        return Form(items)



from distutils.core import setup
setup(
  name = 'yaat',
  packages = ['yaat', 'yaat.middleware'],
  version = '0.1.0',
  license='LGPL',
  description = 'Yet another ASGI toolkit',
  author = 'Khant',
  author_email = 'contact@khant.dev',
  url = 'https://github.com/yaat-project/yaat',
  download_url = 'https://github.com/yaat-project/yaat/releases/download/v0.1.0/yaat-0.1.0.tar.gz',
  keywords = ['Asynchronous', 'Web framework'],
  install_requires=[
    "Jinja2",
    "aiofiles",
    "httpx",
    "parse",
    "python-multipart",
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)



from jinja2 import Environment, FileSystemLoader
import os

from .responses import HTMLResponse


class BaseTemplates:
    """Skeleon class to be inherited for custom template engine"""

    def get(self, *args, **kwargs) -> bytes:
        raise NotImplementedError

    def TemplateResponse(self, *args, **kwargs) -> HTMLResponse:
        raise NotImplementedError


class Jinja2Templates(BaseTemplates):
    def __init__(self, directory: str):
        self.directory = directory

    @property
    def directory(self) -> Environment:
        return self.__directory

    @directory.setter
    def directory(self, directory: str) -> None:
        abspath = os.path.abspath(directory)

        # if path not exists, raise error
        if not os.path.exists(abspath):
            raise FileNotFoundError(f"Directory {abspath} does not exists.")

        self.__directory = Environment(
            loader=FileSystemLoader(abspath)
        )

    def get(self, template_name: str, context: dict = None) -> bytes:
        if context is None:
            context = {}

        return self.directory.get_template(template_name).render(**context).encode()

    def TemplateResponse(
        self,
        template_name: str,
        context: dict = None,
        status_code: int = 200,
        headers: dict = None,
    ) -> HTMLResponse:
        template_data = self.get(template_name, context)
        return HTMLResponse(
            status_code=status_code,
            headers=headers,
            content=template_data
        )
